@@ -53,8 +53,8 @@
 #define READBUF_SIZE 4096
 #define WRITEBUF_SIZE 4096
 #define PIPEBUF_SIZE  4096
-#define MAX_READERS 20
-#define MAX_WRITERS 20
+#define MAX_READERS 40
+#define MAX_WRITERS 40
 
 static struct
 {
@@ -389,9 +389,12 @@ find_reader (int fd, int start_it)
   if (i != reader_table_size)
     {
       rd = create_reader (fd_to_handle (fd));
-      reader_table[i].fd = fd;
-      reader_table[i].context = rd;
-      reader_table[i].used = 1;
+      if (rd)
+	{
+	  reader_table[i].fd = fd;
+	  reader_table[i].context = rd;
+	  reader_table[i].used = 1;
+	}
     }
 
   UNLOCK (reader_table_lock);
@@ -692,9 +695,12 @@ find_writer (int fd, int start_it)
   if (i != writer_table_size)
     {
       wt = create_writer (fd_to_handle (fd));
-      writer_table[i].fd = fd;
-      writer_table[i].context = wt; 
-      writer_table[i].used = 1;
+      if (wt)
+	{
+	  writer_table[i].fd = fd;
+	  writer_table[i].context = wt; 
+	  writer_table[i].used = 1;
+	}
     }
 
   UNLOCK (writer_table_lock);
@@ -733,7 +739,7 @@ _gpgme_io_write (int fd, const void *buffer, size_t count)
   if (count == 0)
     return TRACE_SYSRES (0);
 
-  ctx = find_writer (fd, 1);
+  ctx = find_writer (fd, 0);
   if (!ctx)
     return TRACE_SYSRES (-1);
 
@@ -826,6 +832,7 @@ _gpgme_io_pipe (int filedes[2], int inherit_idx)
   /* Make one end inheritable.  */
   if (inherit_idx == 0)
     {
+      struct writer_context_s *ctx;
       HANDLE hd;
       if (!DuplicateHandle (GetCurrentProcess(), rh,
 			    GetCurrentProcess(), &hd, 0,
@@ -841,9 +848,22 @@ _gpgme_io_pipe (int filedes[2], int inherit_idx)
         }
       CloseHandle (rh);
       rh = hd;
+
+      ctx = find_writer (handle_to_fd (wh), 0);
+      assert (ctx == NULL);
+      ctx = find_writer (handle_to_fd (wh), 1);
+      if (!ctx)
+	{
+	  CloseHandle (rh);
+	  CloseHandle (wh);
+	  /* FIXME: Should translate the error code.  */
+	  errno = EIO;
+	  return TRACE_SYSRES (-1);
+	}
     }
   else if (inherit_idx == 1)
     {
+      struct reader_context_s *ctx;
       HANDLE hd;
       if (!DuplicateHandle( GetCurrentProcess(), wh,
 			    GetCurrentProcess(), &hd, 0,
@@ -859,6 +879,18 @@ _gpgme_io_pipe (int filedes[2], int inherit_idx)
         }
       CloseHandle (wh);
       wh = hd;
+
+      ctx = find_reader (handle_to_fd (rh), 0);
+      assert (ctx == NULL);
+      ctx = find_reader (handle_to_fd (rh), 1);
+      if (!ctx)
+	{
+	  CloseHandle (rh);
+	  CloseHandle (wh);
+	  /* FIXME: Should translate the error code.  */
+	  errno = EIO;
+	  return TRACE_SYSRES (-1);
+	}
     }
   
   filedes[0] = handle_to_fd (rh);
@@ -1009,7 +1041,7 @@ build_commandline (char **argv)
 
 
 int
-_gpgme_io_spawn (const char *path, char *const argv[],
+_gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
 		 struct spawn_fd_item_s *fd_list, pid_t *r_pid)
 {
   SECURITY_ATTRIBUTES sec_attr;
@@ -1021,8 +1053,8 @@ _gpgme_io_spawn (const char *path, char *const argv[],
       0          /* returns tid */
     };
   STARTUPINFO si;
-  int cr_flags = CREATE_DEFAULT_ERROR_MODE
-    | GetPriorityClass (GetCurrentProcess ());
+  int cr_flags = (CREATE_DEFAULT_ERROR_MODE
+                  | GetPriorityClass (GetCurrentProcess ()));
   int i;
   char **args;
   char *arg_string;
@@ -1104,6 +1136,9 @@ _gpgme_io_spawn (const char *path, char *const argv[],
 
   free (arg_string);
 
+  if (flags & IOSPAWN_FLAG_ALLOW_SET_FG)
+    _gpgme_allow_set_foreground_window ((pid_t)pi.dwProcessId);
+
   /* Insert the inherited handles.  */
   for (i = 0; fd_list[i].fd != -1; i++)
     {
@@ -1139,14 +1174,16 @@ _gpgme_io_spawn (const char *path, char *const argv[],
        notation: "0xFEDCBA9876543210" with an extra white space after
        every quadruplet.  10*(19*4 + 1) - 1 = 769.  This plans ahead
        for a time when a HANDLE is 64 bit.  */
-#define BUFFER_MAX 800
+#define BUFFER_MAX 810
     char line[BUFFER_MAX + 1];
     int res;
     int written;
     size_t len;
 
-    line[0] = '\n';
-    line[1] = '\0';
+    if ((flags & IOSPAWN_FLAG_ALLOW_SET_FG))
+      strcpy (line, "~1 \n");
+    else
+      strcpy (line, "\n");
     for (i = 0; fd_list[i].fd != -1; i++)
       {
 	/* Strip the newline.  */
@@ -1181,6 +1218,7 @@ _gpgme_io_spawn (const char *path, char *const argv[],
   
   if (r_pid)
     *r_pid = (pid_t)pi.dwProcessId;
+
   
   if (ResumeThread (pi.hThread) < 0)
     TRACE_LOG1 ("ResumeThread failed: ec=%d", (int) GetLastError ());
@@ -1242,7 +1280,7 @@ _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock)
 	{
 	  if (fds[i].for_read)
 	    {
-	      struct reader_context_s *ctx = find_reader (fds[i].fd,1);
+	      struct reader_context_s *ctx = find_reader (fds[i].fd,0);
 	      
 	      if (!ctx)
 		TRACE_LOG1 ("error: no reader for FD 0x%x (ignored)",
@@ -1265,7 +1303,7 @@ _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock)
             }
 	  else if (fds[i].for_write)
 	    {
-	      struct writer_context_s *ctx = find_writer (fds[i].fd,1);
+	      struct writer_context_s *ctx = find_writer (fds[i].fd,0);
               
 	      if (!ctx)
 		TRACE_LOG1 ("error: no writer for FD 0x%x (ignored)",
@@ -1411,7 +1449,7 @@ _gpgme_io_dup (int fd)
       return TRACE_SYSRES (-1);
     }
 
-  rd_ctx = find_reader (fd, 1);
+  rd_ctx = find_reader (fd, 0);
   if (rd_ctx)
     {
       /* No need for locking, as the only races are against the reader
@@ -1430,7 +1468,7 @@ _gpgme_io_dup (int fd)
       UNLOCK (reader_table_lock);
     }
 
-  wt_ctx = find_writer (fd, 1);
+  wt_ctx = find_writer (fd, 0);
   if (wt_ctx)
     {
       /* No need for locking, as the only races are against the writer
@@ -1469,3 +1507,62 @@ gpgme_get_fdptr (int fd)
 {
   return NULL;
 }
+
+
+static int
+wsa2errno (int err)
+{
+  switch (err)
+    {
+    case WSAENOTSOCK:
+      return EINVAL;
+    case WSAEWOULDBLOCK:
+      return EAGAIN;
+    case ERROR_BROKEN_PIPE:
+      return EPIPE;
+    case WSANOTINITIALISED:
+      return ENOSYS;
+    default:
+      return EIO;
+    }
+}
+
+
+int
+_gpgme_io_socket (int domain, int type, int proto)
+{
+  int res;
+
+  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_socket", domain,
+	      "type=%i, protp=%i", type, proto);
+
+  res = socket (domain, type, proto);
+  if (res == INVALID_SOCKET)
+    {
+      errno = wsa2errno (WSAGetLastError ());
+      return TRACE_SYSRES (-1);
+    }
+
+  TRACE_SUC1 ("socket=0x%x", res);
+  
+  return res;
+}
+
+
+int
+_gpgme_io_connect (int fd, struct sockaddr *addr, int addrlen)
+{
+  int res;
+
+  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_connect", fd,
+	      "addr=%p, addrlen=%i", addr, addrlen);
+
+  res = connect (fd, addr, addrlen);
+  if (!res)
+    {
+      errno = wsa2errno (WSAGetLastError ());
+      return TRACE_SYSRES (-1);
+    }
+
+  return TRACE_SUC ();
+}
